import logging
from typing import Dict

import pandas as pd

from pydelling.paraview_processor.filters import VtkFilter, \
    BaseFilter, CalculatorFilter, IntegrateVariablesFilter, PlotOverLineFilter, XDMFFilter, CellDataToPointDataFilter, \
    ClipFilter, StreamTracerWithCustomSourceFilter, SaveDataFilter, AppendArcLengthFilter, TableToPointsFilter, \
    CsvReaderFilter

logger = logging.getLogger(__name__)
try:
    from paraview.simple import *
    from paraview.vtk.numpy_interface import dataset_adapter as dsa
    from paraview import servermanager as sm
except:
    logger.warning("Paraview python implementation is not properly set-up")

class ParaviewProcessor:
    """
    This class provides the framework to read data from a VTK file and do different postprocessing steps
    """
    current_array: None
   # calculator: Calculator
    pipeline: Dict[str, BaseFilter] = {}
    vtk_data_counter: int = 0
    xdmf_data_counter: int = 0
    csv_data_counter: int = 0

    def add_vtk_file(self, path, name=None) -> VtkFilter:
        """
        Reads a given vtk file. This is done by adding an instance of the LegacyVTKReader class to the pipeline.
        Args:
            filename: The path of the vtk file.
            name: A custom name to be given in the pipeline.
        Returns:
            The created LegacyVTKReader instance
        """
        pipeline_name = name if name else f"vtk_data_{VtkFilter.counter}"
        self.vtk_data_counter += 1
        vtk_filter = VtkFilter(filename=str(path), name=pipeline_name)
        self.pipeline[pipeline_name] = vtk_filter
        logger.info(f"Added VTK file {path} as {vtk_filter.name} object to Paraview processor")
        return vtk_filter

    def add_csv_file(self, path, name=None, coordinate_labels=("x", "y", "z")) -> CsvReaderFilter:
        """
        Reads a given csv file. A TableToPoints filter is applied to the read csv file
        Args:
            filename: The path of the vtk file.
            name: A custom name to be given in the pipeline.
            coordinate_labels: The names of the x, y and z variables used on the csv file
        Returns:
            The TableToPoints filter containing the csv data
        """
        pipeline_name = name if name else f"vtk_data_{VtkFilter.counter}"
        self.csv_data_counter += 1
        csv_reader = CsvReaderFilter(filename=str(path), name=pipeline_name, coordinate_labels=coordinate_labels)
        self.pipeline[pipeline_name] = csv_reader
        logger.info(f"Added CSV file {path} as {csv_reader.name} object to Paraview processor")
        return csv_reader

    def add_xdmf_file(self, path, name=None) -> XDMFFilter:
        """
        Reads a given xdmf file.
        Args:
            filename: The path of the xdmf file.
            name: A custom name to be given in the pipeline.
        Returns:
            The created XDMFReader instance
        """
        pipeline_name = name if name else f"xdmf_data_{XDMFFilter.counter}"
        self.vtk_data_counter += 1
        pv_filter = XDMFFilter(filename=str(path), name=pipeline_name)
        self.pipeline[pipeline_name] = pv_filter
        logger.info(f"Added XDMF file {path} as {pv_filter.name} object to Paraview processor")
        return pv_filter

    def add_calculator(self, input_filter, function='', name=None, output_array_name='Results', *args, **kwargs) -> CalculatorFilter:
        """
        Adds a calculator filter to a dataset
        Returns:
            The Calculator object
        """
        pipeline_name = name if name else f"calculator_{CalculatorFilter.counter}"
        calculator_filter = CalculatorFilter(input_filter=self.process_input_filter(filter=input_filter),
                                           function=function,
                                           name=pipeline_name,
                                           output_array_name=output_array_name, *args, **kwargs)
        self.pipeline[pipeline_name] = calculator_filter
        logger.info(f"Added calculator filter based on {self.get_input_object_name(input_filter)} as {calculator_filter.name} object to Paraview processor")
        return calculator_filter

    def add_cell_data_to_point_data(self, input_filter, name=None) -> CellDataToPointDataFilter:
        """
        Adds a cell data to point data filter to a dataset
        Returns:
            The CellDataToPointDataFilter object
        """
        pipeline_name = name if name else f"cell_data_to_point_data{CellDataToPointDataFilter.counter}"
        pv_filter = CellDataToPointDataFilter(input_filter=self.process_input_filter(filter=input_filter),
                                           name=pipeline_name,
                                                      )
        self.pipeline[pipeline_name] = pv_filter
        logger.info(f"Added cell_data_to_point_data filter based on {self.get_input_object_name(input_filter)} as {pv_filter.name} object to Paraview processor")
        return pv_filter

    def add_clip(self, input_filter, name=None, *args, **kwargs) -> ClipFilter:
        """
        Adds a clip filter to a dataset
        Returns:
            The ClipFilter object
        """
        pipeline_name = name if name else f"clip_{ClipFilter.counter}"
        pv_filter = ClipFilter(input_filter=self.process_input_filter(filter=input_filter),
                                           name=pipeline_name, *args, **kwargs
                              )
        self.pipeline[pipeline_name] = pv_filter
        logger.info(f"Added clip filter based on {self.get_input_object_name(input_filter)} as {pv_filter.name} object to Paraview processor")
        return pv_filter

    def add_table_to_points(self, path, name=None) -> TableToPointsFilter:
        """
        Adds a table to points filter to a dataset
        Returns:
            The TableToPointsFilter object
        """
        pipeline_name = name if name else f"table_to_points_{TableToPointsFilter.counter}"
        pv_filter = TableToPointsFilter(filename=str(path), name=pipeline_name)
        self.pipeline[pipeline_name] = pv_filter
        logger.info(f"Added table to points filter based on {path} as {pv_filter.name} object to Paraview processor")
        return pv_filter


    def add_stream_tracer_with_custom_source(self, input_filter, seed_source, name=None) -> StreamTracerWithCustomSourceFilter:
        """
        Adds a stream tracer filter to a custom source
        Returns:
            The StreamTracerWithCustomSourceFilter object
        """
        pipeline_name = name if name else f"stream_tracer_with_custom_source_{StreamTracerWithCustomSourceFilter.counter}"
        pv_filter = StreamTracerWithCustomSourceFilter(input_filter=self.process_input_filter(filter=input_filter),
                                           seed_source=self.process_input_filter(filter=seed_source),
                                           name=pipeline_name,
                                                        )
        self.pipeline[pipeline_name] = pv_filter
        logger.info(f"Added stream tracer filter based on {self.get_input_object_name(input_filter)} as {pv_filter.name} object to Paraview processor")
        return pv_filter

    def add_append_arc_length(self, input_filter, name=None) -> AppendArcLengthFilter:
        """
        Adds an append arc-length filter to a dataset
        Returns:
            The AppendArcLengthFilter object
        """
        pipeline_name = name if name else f"append_arc_length_{AppendArcLengthFilter.counter}"
        pv_filter = AppendArcLengthFilter(input_filter=self.process_input_filter(filter=input_filter),
                                           name=pipeline_name
                              )
        self.pipeline[pipeline_name] = pv_filter
        logger.info(f"Added append arc length filter based on {self.get_input_object_name(input_filter)} as {pv_filter.name} object to Paraview processor")
        return pv_filter

    def add_save_data(self, path, proxy, PointDataArrays, CellDataArrays, name=None) -> SaveDataFilter:
        """
        Writes a csv file.
        Args:
            filename: The path of the xdmf file.
            name: A custom name to be given in the pipeline.
        Returns:
            The created SaveDataFilter object
        """
        pipeline_name = name if name else f"save_data_{SaveDataFilter.counter}"
        pv_filter = SaveDataFilter(filename=str(path), proxy=self.process_input_filter(filter=proxy),
                                   point_data_arrays=PointDataArrays, cell_data_arrays=CellDataArrays,
                                   name=pipeline_name)
        self.pipeline[pipeline_name] = pv_filter
        logger.info(f"Added csv file based on {self.get_input_object_name(proxy)} as {pv_filter.name} object to Paraview processor")
        return pv_filter


    def add_integrate_variables(self, input_filter, name=None, divide_cell_data_by_volume=False) -> IntegrateVariablesFilter:
        """
        Adds the integrate_variables filter to a dataset
        Returns:
            An IntegrateVariablesFilter object
        """
        pipeline_name = name if name else f"integrate_variables_{IntegrateVariablesFilter.counter}"
        integrate_variables_filter = IntegrateVariablesFilter(input_filter=self.process_input_filter(filter=input_filter),
                                                              name=pipeline_name,
                                                              divide_cell_data_by_volume=divide_cell_data_by_volume
                                                              )
        self.pipeline[pipeline_name] = integrate_variables_filter
        logger.info(
            f"Added integrate_variables filter based on {self.get_input_object_name(input_filter)} as {integrate_variables_filter.name} object to Paraview processor")
        return integrate_variables_filter

    def add_plot_over_line(self, input_filter, name=None, point_1=None, point_2=None, line_resolution=None) -> PlotOverLineFilter:
        """
        Adds the plot_over_line filter to a dataset
        Returns:
            A [PlotOverLineFilter][pydelling/paraview_processor/filters/PlotOverLineFilter.py] object
        """
        pipeline_name = name if name else f"plot_over_line_{PlotOverLineFilter.counter}"
        plot_over_line_filter = PlotOverLineFilter(input_filter=self.process_input_filter(filter=input_filter),
                                                   name=pipeline_name,
                                                   point_1=point_1,
                                                   point_2=point_2,
                                                   n_line=line_resolution,
                                                   )
        self.pipeline[pipeline_name] = plot_over_line_filter
        logger.info(
            f"Added plot_over_line_filter filter based on {self.get_input_object_name(input_filter)} as {plot_over_line_filter.name} object to Paraview processor")
        return plot_over_line_filter

    # Utility methods
    def plot_over_z_given_xy_point(self, dataset: BaseFilter, x_point, y_point, line_resolution=None) -> PlotOverLineFilter:
        """
        This method interpolates the dataset over a line going in the z-axis given XY coordinates in space
        Args:
            x_point: X coordinate
            y_point: Y coordinate

        Returns:
            A plot_over_line object
        """
        # Compute z_min and z_max values of the dataset
        point_1 = [x_point, y_point, dataset.z_min]
        point_2 = [x_point, y_point, dataset.z_max]
        plot_over_line_filter = PlotOverLineFilter(input_filter=self.process_input_filter(filter=dataset),
                                                   name="plot_over_z_given_xy_point",
                                                   point_1=point_1,
                                                   point_2=point_2,
                                                   n_line=line_resolution,
                                                   )
        return plot_over_line_filter

    def aperture_given_xy_point(self, dataset: BaseFilter, x_point, y_point, variable=None, target_value=1.0, threshold=0.45, line_resolution=100) -> float:
        """
        This method computes the aperture at a given position in the XY plane.
        Args:
            x_point: X coordinate
            y_point: Y coordinate
            variable: The variable to consider that indicates the aperture
            target_value: Value of the variable when the fracture is open
            threshold: Value of the maximum variation from the target_value |line[variable] - target_value| < threshold is assumed
            line_resolution: Resolution of the line interpolation that is being used
        Returns:
            The value of the aperture at a given point
        """
        line_interpolation = self.plot_over_z_given_xy_point(x_point=x_point,
                                                             y_point=y_point,
                                                             line_resolution=line_resolution,
                                                             dataset=dataset,
                                                             )
        line_interpolation_point_data = line_interpolation.point_data
        dataset_mesh_points = dataset.mesh_points
        if variable:
            # If a target variable is defined,
            line_interpolation = line_interpolation_point_data[abs(line_interpolation_point_data[variable] - target_value) < threshold]
            z_points: pd.Series = dataset_mesh_points.iloc[line_interpolation.index]["z"]
            if len(z_points) > 0:
                aperture = abs(z_points.max() - z_points.min())
            else:
                aperture = 0.0
            return aperture

        if not variable:
            # Calculate directly the aperture
            z_points: pd.Series = dataset_mesh_points.iloc[line_interpolation_point_data.index]["z"]
            if len(z_points) > 0:
                aperture = abs(z_points.max() - z_points.min())
            else:
                aperture = 0.0
            return aperture




    def print_pipeline(self) -> str:
        """
        Creates a simple representation of the current working filters and objects
        Returns:
            A string containing the pipeline structure
        """
        identation_level: int = 0
        print_string: str = f""
        print_string += "Paraview filters:\n"
        identation_level += 1
        for pipeline_element in self.pipeline:
            # if type(self.pipeline[pipeline_element]) == dict:
            #     print_string = self.print_pipeline_block(pipeline_dict=self.pipeline[pipeline_element],
            #                               starting_identation_level=identation_level+1,
            #                               output_string=print_string)
            #     continue
            identation = "\t"*identation_level
            print_string += f"{identation}- {pipeline_element} [{self.pipeline[pipeline_element].filter_type}]\n"
        print(print_string)
        return print_string

    def print_pipeline_block(self, pipeline_dict: Dict, output_string:str,  starting_identation_level: int = 0):
        for pipeline_element in pipeline_dict:
            if type(pipeline_dict[pipeline_element]) == dict:
                output_string = self.print_pipeline_block(pipeline_dict=pipeline_dict[pipeline_element],
                                          starting_identation_level=starting_identation_level+1,
                                          output_string=output_string)
                continue
            identation = "\t"*starting_identation_level
            output_string += f"{identation}- {pipeline_element}\n"
        return output_string

    def get_object(self, name):
        """
        Get a given object from the pipeline
        Args:
            name: The name of the object to obtain

        Returns:
            The requested object
        """
        return self.pipeline[name]

    def process_input_filter(self, filter) -> object:
        """
        Processes the filter object and returns the proper datatype
        Args:
            input_filter: an object refering to an existing filter

        Returns:
            A proper Paraview object
        """
        if type(filter) == str:
            # Assume the filter specifies the name of the pipeline
            return self.pipeline[filter].filter
        else:
            return filter.filter

    @staticmethod
    def get_input_object_name(filter) -> str:
        """
        Gets the name of a given filter
        Args:
            filter: filter to be processed

        Returns:
            String containing the filter's name
        """
        if type(filter) == str:
            return filter
        else:
            return filter.name




    def __repr__(self):
        return self.print_pipeline()



"""
Centroid file reader
"""
import logging
import pickle

import numpy as np
import pandas as pd
from pandas.core.groupby import DataFrameGroupBy
from tqdm import tqdm

from pydelling.config import config
from .BaseReader import BaseReader

logger = logging.getLogger(__name__)


class StreamlineReader(BaseReader):
    data: pd.DataFrame
    raw_data: pd.DataFrame
    stream_data: DataFrameGroupBy
    is_aperture_zero: bool = False

    def __init__(self, filename, header=False):
        self.header = None
        super().__init__(filename,
                         header=header)

    def read_file(self, opened_file):
        """
        This method reads an opened file an sets the `self` variables properly to be used throuhgout the class
        Args:
            opened_file: the object containing the opened file
        """
        logger.info(f"Reading streamlines from {self.filename}")
        self.data = pd.read_csv(opened_file)
        self.raw_data = self.data.copy()  # Copy of original data
        self.data = self.data.rename(columns={"Points:0": "x",
                                              "Points:1": "y",
                                              "Points:2": "z",
                                              }
                                     )
        self.data = self.data.apply(pd.to_numeric, args=("coerce",))
        self.stream_data: DataFrameGroupBy = self.data.groupby("SeedIds")

    def compute_arrival_times(self,
                              reason_of_termination=None,
                              min_x=None,
                              min_y=None,
                              ) -> pd.Series:
        """
        This method computes the arrival times of the streamlines

        Args:
            reason_of_termination: Filter by the Paraview tag that defines the status of the streamlines
            min_x: Filters the streamlines with their max x < min_x

        Returns:
             A pd.Series object containing the arrival times of the streamlines

        """
        logger.info("Computing arrival times of the streamlines")
        filtered_streamlines = self.filter_streamlines(reason_of_termination=reason_of_termination, min_x=min_x, min_y=min_y)
        temp_series: pd.Series = filtered_streamlines.max()["IntegrationTime"]
        return temp_series

    def compute_arrival_times_per_material(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the arrival times of the streamlines for a particular material
        Returns:
             A pd.Series object containing the arrival times of the streamlines
        """
        logger.info("Computing arrival times of the streamlines per material")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)

        temp_df['Material ID'] = (temp_df['Material ID'] + 0.45).apply(np.floor)
        temp_series: pd.Series = temp_df.groupby(["Material ID", "SeedIds"]).max()['IntegrationTime']
        temp_series = temp_series.reset_index()

        return temp_series



    def compute_arrival_times_per_material_paula(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the arrival times of the streamlines for a particular material
        Returns:
             A pd.Series object containing the arrival times of the streamlines
        """
        logger.info("Computing arrival times of the streamlines per material")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data.copy()
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)

        # temp_df['Material ID'] = temp_df['Material ID'].apply(np.ceil)
        temp_df['Material ID'] = (temp_df['Material ID'] + 0.45).apply(np.floor)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1, 1.5), 1)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1.5, 2.5), 2)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(2.5, 3.5), 3)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(3.5, 4.5), 4)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(4.5, 5), 5)
        temp_series: pd.Series = temp_df.groupby(["Material ID", "SeedIds"]).max()
        temp_series = temp_series.reset_index()

        # AUTOMATIC
        dic = {}
        dic_group = {}
        for group in temp_series.groupby('Material ID').groups:
            ngroup = temp_series.groupby('Material ID').get_group(group)['IntegrationTime']
            dic[group] = ngroup
            dic_group[group] = ''

        b = True
        for key in sorted(dic.keys(), reverse=True):
            if b:
                add = dic[key]
                dic_group[key] = add
                b = False
            else:
                add = pd.concat([add, dic[key]])
                dic_group[key] = add

        return temp_series, dic_group

    def compute_initial_velocities(self,
                                   reason_of_termination = None,
                                   normalize=True,
                                   index_df: pd.Series=None,
                                   ):
        """
        This method computes the initial velocities each streamlines 'sees' at the beginning, it can be used to normalize them later on

        Args:
            normalized: Parameter controlling weather the output vector should be normalized by dividing by the maximum velocity valuefdh

        Returns: a vector containing the initial velocities of the streamlines
        """
        logger.info("Computing initial velocities of the streamlines")
        temp_df = self.data
        if index_df is not None:
            seed_ids = index_df["SeedIds"]
            temp_df = temp_df[temp_df["SeedIds"].isin(seed_ids)]
        temp_df = temp_df.groupby("SeedIds").first()
        temp_series = temp_df["U:0"]
        if normalize:
            temp_series /= temp_series.sum()
        temp_series[temp_series < 0.0] = 0.0
        return temp_series

    def compute_length_streamlines(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the length of the streamlines
        Returns:
             A pd.Series object containing the length of the streamlines
        """
        logger.info("Computing length of the streamlines")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)
        temp_series: pd.Series = temp_df.groupby("SeedIds").max()["arc_length"]
        return temp_series

    def compute_length_streamlines_per_material(self, reason_of_termination=None) -> pd.Series:
        """
        This method computes the length of the streamlines for a particular material
        Returns:
             A pd.Series object containing the length of the streamlines
        """
        logger.info("Computing length of the streamlines per material")
        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.reason_of_termination
        temp_df = self.stream_data
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x["ReasonForTermination"].max() == reason_of_termination)

        # temp_df['Material ID'] = temp_df['Material ID'].apply(np.ceil)
        temp_df['Material ID'] = (temp_df['Material ID'] + 0.45).apply(np.floor)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1, 1.5), 1)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(1.5, 2.5), 2)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(2.5, 3.5), 3)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(3.5, 4.5), 4)
        # temp_df['Material ID'] = temp_df['Material ID'].mask(temp_df['Material ID'].between(4.5, 5), 5)

        temp_series: pd.Series = temp_df.groupby(["Material ID", "SeedIds"]).max()
        temp_series = temp_series.reset_index()

        # AUTOMATIC
        # dic = {}
        # for group in temp_series.groupby('Material ID').groups:
        #     ngroup_series = temp_series.groupby('Material ID').get_group(group)['arc_length']
            # temp_series = temp_series.groupby('Material ID').get_group(group)['arc_length']
            # dic[group] = ngroup

        return temp_series #, dic
        # return ngroup_series

    def compute_beta(self, aperture_field: str = None) -> pd.Series:
        """
        This method computes beta values for each streamline
        Returns:
            A pd.Series object containing the streamline info with the beta column added
        """
        self.number_of_zero_apertures = 0
        aperture_field_file = aperture_field if aperture_field else config.beta_integrator.aperture_field_file if config.beta_integrator.aperture_field_file else None
        assert aperture_field_file, "Define a file containing an aperture field matrix"
        # Read aperture field
        with open(aperture_field_file, "rb") as opened_file:
            aperture_field = pickle.load(opened_file)
        aperture_field = self.fix_aperture_field(aperture_field)
        logger.info(f"Aperture field has been loaded from {aperture_field_file}")
        logger.info("Computing beta values for the streamlines")
        self.data["beta"] = 0.0
        # Filter streamlines
        temp_df = self.filter_streamlines()

        for stream in tqdm(temp_df.groups):
            # if random.random() > 0.01:
            #     continue
            stream_data = self.stream_data.get_group(stream)
            index_group = stream_data.index
            stream_beta_value = self.integrate_beta(stream=stream_data, aperture_field=aperture_field)
            if stream_beta_value == 0.0:
                continue
            if self.is_aperture_zero:
                self.is_aperture_zero = False
                continue
            else:
                self.data.loc[index_group, "beta"] = stream_beta_value

        return self.data.groupby('SeedIds').max()["beta"]

    def integrate_beta(self, stream: pd.DataFrame, aperture_field: np.ndarray):
        """
        This method integrates a given streamline to compute the value of beta
        Args:
            stream: A pandas DataFrame containing the data of a given streamline
            aperture_field: A matrix containing the information of the aperture field
        Returns:

        """
        beta = 0.0
        aperture_field_nx = aperture_field.shape[1]
        aperture_field_ny = aperture_field.shape[0]
        previous_integration_time = 0.0
        previous_aperture = 0.0
        for index, fragment in stream.iterrows():
            # aperture = aperture_from_a_xy_point(x_point=)
            # Nearest neighbour
            index_x = int(np.floor(fragment["x"] / config.beta_integrator.dimension_x * aperture_field_nx))
            index_y = int(np.floor(fragment["y"] / config.beta_integrator.dimension_y * aperture_field_ny))
            index_row = index_y
            index_column = index_x
            if index_row == aperture_field_ny:
                index_row -= 1
            if index_column == aperture_field_nx:
                index_column -= 1

            # Compute tau
            tau = fragment["IntegrationTime"] - previous_integration_time
            previous_integration_time = fragment["IntegrationTime"]
            # if tau < 0.0:
            #     print(tau)

            # Calculate aperture
            aperture = aperture_field[index_row, index_column]
            # if fragment["x"] < 0.001:
            #     print(f"Aperture: {aperture}, x: {fragment['x']} y: {fragment['y']} index_x: {index_x}, index_y: {index_y}")
            if aperture == 0.0:
                aperture = previous_aperture
                continue
            previous_aperture = aperture
            # print(f"Tau: {tau / (365 * 24)} h, Aperture: {aperture} m, Fragment beta: {2 * tau / aperture / (365 * 24 * 3600)}, Cummulated beta: {beta}")
            beta += 2 * tau / aperture / (365 * 24 * 3600)
        # print(f"Stream Finishes")
        # print(f"Computed beta: {beta}")
        return beta

    def get_data(self) -> np.ndarray:
        """
        Outputs the data
        :return: np.ndarray object containing centroid information and variable output
        """
        return self.data.values

    def dump_to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        # self.data.to_csv(output_file, delimiter=delimiter)
        self.data.to_csv(output_file)
        print(f"The data has been properly exported to the {output_file} file")

    def to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        # self.data.to_csv(output_file, delimiter=delimiter)
        self.data.to_csv(output_file)
        print(f"The data has been properly exported to the {output_file} file")

    def filter_streamlines(self,
                           reason_of_termination=None,
                           min_x=None,
                           min_y=None,
                           ) -> DataFrameGroupBy:


        reason_of_termination = reason_of_termination if reason_of_termination else config.streamline_reader.filter.reason_of_termination if config.streamline_reader.filter.reason_of_termination else None
        min_x = min_x if min_x else config.streamline_reader.filter.min_x if config.streamline_reader.filter.min_x else None
        logger.info("Filtering streamlines")
        temp_df = self.data.groupby("SeedIds")
        initial_number_of_streamlines = len(temp_df.groups)
        if reason_of_termination:
            temp_df = temp_df.filter(lambda x: x.max()["ReasonForTermination"] == float(reason_of_termination))
            temp_df = temp_df.groupby("SeedIds")
            logger.info(f"{initial_number_of_streamlines - len(temp_df.groups)} streamlines have been filtered due to 'Reason of termination' filtering (Keeping {len(temp_df.groups) / initial_number_of_streamlines * 100: 2.1f}% of total)")
            # initial_number_of_streamlines = temp_df.shape[0]
        if min_x is not None:
            temp_df = temp_df.filter(lambda x: x.max()["x"] > min_x)
            temp_df = temp_df.groupby("SeedIds")
            logger.info(f"{initial_number_of_streamlines - len(temp_df.groups)} streamlines have been filtered due to 'min_x' filtering (Keeping {len(temp_df.groups) / initial_number_of_streamlines * 100:2.1f}% of total)")

        if min_y is not None:
            temp_df = temp_df.filter(lambda x: x.max()["y"] > min_y)
            temp_df = temp_df.groupby("SeedIds")
            logger.info(f"{initial_number_of_streamlines - len(temp_df.groups)} streamlines have been filtered due to 'min_y' filtering (Keeping {len(temp_df.groups) / initial_number_of_streamlines * 100:2.1f}% of total)")
        return temp_df

    @staticmethod
    def fix_aperture_field(aperture_matrix):
        aperture_matrix[0, :] = aperture_matrix[1, :]
        aperture_matrix[:, 0] = aperture_matrix[:, 1]
        aperture_matrix[aperture_matrix.shape[0] - 1, :] = aperture_matrix[aperture_matrix.shape[0] - 2, :]
        aperture_matrix[:, aperture_matrix.shape[1] - 1] = aperture_matrix[:, aperture_matrix.shape[1] - 2]
        return aperture_matrix

    def integrate_variable_within_materials(self, variable,
                                            material_names=None,
                                            add_variable_name_to_output=True,
                                            output_variable_name = None,
                                            *args,
                                            **kwargs
                                            ):
        logger.info(f'Integrating variable {variable} within the materials for all the streamlines')
        output_list = []
        for streamline in self.stream_data:
            output_list.append(self._integrate_variable_within_materials_single(streamline=streamline[1],
                                                           variable=variable,
                                                           *args,
                                                           **kwargs)
                             )
        output_df = pd.DataFrame(output_list)
        if material_names:
            if add_variable_name_to_output:
                if not output_variable_name:
                    output_variable_name = variable
                output_column_names = [f"{material_names[column]}-{output_variable_name}" for column in output_df.columns]
            else:
                output_column_names = [material_names[column] for column in output_df.columns]
            output_df.columns = output_column_names
        return output_df


    @staticmethod
    def _integrate_variable_within_materials_single(streamline: pd.DataFrame,
                                            variable: str,
                                            categorize_materials: bool=True,
                                            material_id_column: str='Material ID',
                                            material_id_offset: float = 0.45,
                                            ) -> dict:
        """
        This method implements an algorithm to integrate a given variable along the different materials of a streamline
        Args:
            variable: variable to integrate

        Returns:
            pd.Dataframe containing the materials and the integrated variable
        """
        streamline = streamline.reset_index()
        if categorize_materials:
            streamline[material_id_column] = (streamline[material_id_column] + material_id_offset).apply(np.floor)
        material_groups = streamline.groupby(material_id_column).groups
        output_dict = {material: 0.0 for material in material_groups.keys()}
        var_current = None
        for idx, segment in streamline.iterrows():
            if idx == 0:
                # Initial segment
                output_dict[segment[material_id_column]] += segment[variable]
                var_current = segment[variable]
                continue
            if var_current == None:
                raise ValueError('There was an error when processing the initial segment')
            output_dict[segment[material_id_column]] += segment[variable] - var_current
            var_current = segment[variable]
        return output_dict


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydelling.readers.iGPReader import iGPReader
import logging

logger = logging.getLogger(__name__)


class RegionOperations:
    boundaries: dict
    region_dict: dict
    # Add iGPReader class methods to this class namespace

    def divide_topography_by_z(self: iGPReader, z: float,
                               region_name: str,
                               name_below='below',
                               name_above='above',
                               ):
        """
        Divide the topography into two regions, one above and one below a given z coordinate.

        Args:
            z: The z coordinate to divide the topography by.
            region_name: The name of the region to divide.
            name_below: The name of the region below the z coordinate.
            name_above: The name of the region above the z coordinate.

        Returns:
            None
        """
        boundary_faces = self.get_boundary_faces(region_name)
        name_below = name_below if name_below != 'below' else f"{region_name}_below"
        name_above = name_above if name_above != 'above' else f"{region_name}_above"
        logger.info(f'Dividing topography by z={z} into regions {name_below} and {name_above}')

        new_boundaries = {name_below: [],
                          name_above: []}

        for face_idx, face in enumerate(boundary_faces):
            if face.centroid[2] < z:
                new_boundaries[name_below].append(face_idx)
            else:
                new_boundaries[name_above].append(face_idx)

        # Update the boundaries dict
        for name in new_boundaries:
            self.boundaries[name] = [boundary_faces[idx] for idx in new_boundaries[name]]
        self.boundaries.pop(region_name)

        # Update the region dict
        get_old_region = self.region_dict[region_name]
        # Divide the region elements based on the found idx
        self.region_dict[name_below] = {'elements': get_old_region['elements'][new_boundaries[name_below]],
                                        'centroid_id': get_old_region['centroid_id'][new_boundaries[name_below] ],
                                        'length': len(new_boundaries[name_below])}
        self.region_dict[name_above] = {'elements': get_old_region['elements'][new_boundaries[name_above]],
                                        'centroid_id': get_old_region['centroid_id'][new_boundaries[name_above]],
                                        'length': len(new_boundaries[name_above])}

        # Remove the old region
        self.region_dict.pop(region_name)






import logging
from pathlib import Path
from typing import Dict

import numpy as np
import pandas as pd
from scipy import stats

from pydelling.config import config
from pydelling.readers.iGPReader.io import BaseReader, iGPReader
from pydelling.readers.iGPReader.utils import utils

logger = logging.getLogger(__name__)
import matplotlib.pyplot as plt


class BoreholeReader(BaseReader):
    """This class reads borehole information from csv files and processes it to automatically assign to the mesh"""
    def __init__(self, igp_reader=None, filename=None):
        self.filename = filename if filename else config.data_files.borehole_data if config.data_files.borehole_data else None
        self.igp_reader: iGPReader = igp_reader
        self.boreholes_info = {}

    def run(self):
        self.read()
        self.set_materials_to_boreholes()
        if config.borehole_processing.permeability_from_borehole_data:
            self.set_permeability_from_excel()
        if config.borehole_processing.porosity_from_borehole_data:
            self.set_porosity_from_excel()
        if config.borehole_processing.make_material_plots:
            self.make_material_plots()

    def read(self):
        """Reads the borehole information data from a excel file"""
        if self.filename:
            if not Path(self.filename).suffix == ".xlsx" or Path(self.filename).suffix == ".xls":
                logger.error(f"{self.filename} is not an Excel file")
                raise AssertionError(f"{self.filename} is not an Excel file")
            if config.borehole_processing.anisotropic_borehole_data:
                logger.info('Processing the file as containing anisotropic permeability values')
                self.data = pd.read_excel(self.filename, header=3, sheet_name=None)
                for borehole in self.data:
                    self.data[borehole] = self.data[borehole].iloc[1:]
                    self.data[borehole] = self.data[borehole]
                    self.data[borehole].rename(columns={'Permeability (m^2)': 'kx',
                                                'Unnamed: 3': 'ky',
                                                'Unnamed: 4': 'kz',
                                                }, inplace=True)
                    self.data[borehole]['kmean'] = stats.gmean(np.array([self.data[borehole]['kx'],
                                                                         self.data[borehole]['ky'],
                                                                         self.data[borehole]['kz'],
                                                                         ]).astype(np.float)
                                                               )
            else:
                self.data = pd.read_excel(self.filename, header=3, sheet_name=None)
            _excel_file: Dict[str, pd.DataFrame] = pd.read_excel(self.filename, sheet_name=None)
        for borehole in _excel_file:
            self.boreholes_info[borehole] = {"x": _excel_file[borehole].iloc[0, 1],
                                        "y": _excel_file[borehole].iloc[0, 2],
                                        "z": _excel_file[borehole].iloc[0, 3],
                                             }

    def set_permeability_from_excel(self):
        logger.info(f"Setting the permeability values based on the {self.filename} file")
        for material in self.igp_reader.material_dict:
            # Compute an approximated permeability value for each material
            geometric_mean_boreholes = []
            for borehole in self.data:
                filtered_borehole = self.data[borehole][self.data[borehole]["Material_name"] == material]
                if config.borehole_processing.anisotropic_borehole_data:
                    geometric_mean_boreholes.append(stats.gmean(filtered_borehole["kmean"]))
                else:
                    geometric_mean_boreholes.append(stats.gmean(filtered_borehole["Permeability (m^2)"]))
            material_permeability = stats.gmean(geometric_mean_boreholes)
            self.igp_reader.material_info[material]["permeability"] = material_permeability

    def set_porosity_from_excel(self):
        logger.info(f"Setting the porosity values based on the {self.filename} file")
        for material in self.igp_reader.material_dict:
            # Compute an approximated permeability value for each material
            arithmetic_mean_boreholes = []
            for borehole in self.data:
                filtered_borehole = self.data[borehole][self.data[borehole]["Material_name"] == material]
                arithmetic_mean_boreholes.append(stats.hmean(filtered_borehole["Porosity (-)"]))
            material_porosity = stats.hmean(arithmetic_mean_boreholes)
            self.igp_reader.material_info[material]["porosity"] = material_porosity

    def set_materials_to_boreholes(self):
        """This method should assign a material ID to each depth of the borehole"""
        logger.info("Setting materials to boreholes")
        for borehole in self.boreholes_info:
            self.data[borehole].insert(self.data[borehole].shape[1], "Material_ID", np.zeros(self.data[borehole].shape[0]))
            # create a coordinate dataframe
            _borehole_coordinates = self.data[borehole][["Z (mAOD)", "Material_ID"]]
            _borehole_coordinates.insert(0, "x", self.boreholes_info[borehole]["x"] * np.ones(self.data[borehole].shape[0]))
            _borehole_coordinates.insert(1, "y", self.boreholes_info[borehole]["y"] * np.ones(self.data[borehole].shape[0]))
            _borehole_coordinates.columns = ["x", "y", "z", "Material_ID"]
            logger.debug(f"Calculating material IDs for borehole {borehole}")
            _material_id_vector = utils.get_material_id_from_centroids(igp_reader=self.igp_reader,
                                                                       borehole_coordinates=_borehole_coordinates,
                                                                       )
            self.data[borehole]["Material_ID"] = _material_id_vector
            self.data[borehole]["Material_name"] = [self.igp_reader.material_names[material_id] for material_id in self.data[borehole]["Material_ID"]]

    def make_material_plots(self):
        # print(self.data['Borehole1Data'])
        if config.borehole_processing.permeability_from_borehole_data:
            if config.borehole_processing.anisotropic_borehole_data:
                self.borehole_plot(variable='kx')
                self.borehole_plot(variable='ky')
                self.borehole_plot(variable='kz')
                self.borehole_plot(variable='kmean')
            else:
                self.borehole_plot(variable='Permeability (m^2)')
        if config.borehole_processing.porosity_from_borehole_data:
            self.borehole_plot(variable='Porosity (-)')

    def borehole_plot(self, variable) -> [plt.Figure, plt.Axes]:
        """This method plots the permeability and porosity values taken from the boreholes for each material"""
        logger.info(f'Plotting borehole {variable} data vs "Z (mAOD)"')
        fig: plt.Figure
        ax: plt.Axes
        fig, ax = plt.subplots()
        if config.borehole_processing.make_material_plots.grid:
            ax.grid(zorder=0)
        ax.set_axisbelow(True)
        ax.set_ylabel('Z (mAOD)')
        plot_data = {}
        dictionary_built = False
        for borehole in self.data:
            # Build data structure
            for material in self.data[borehole]['Material_name'].unique():
                if dictionary_built == False:
                    plot_data[material] = {'x': [], 'y': []}
            # Add plot data based on the boreholes
            dictionary_built = True
            for material in self.data[borehole]['Material_name'].unique():
                plot_data[material]['x'].append(self.data[borehole][self.data[borehole]['Material_name'] == material][variable].values.astype(np.float))
                if config.borehole_processing.make_material_plots.plot_cross_correlation:
                    cross_correlation_var = config.borehole_processing.make_material_plots.plot_cross_correlation.variable
                    plot_data[material]['y'].append(self.data[borehole][self.data[borehole]['Material_name'] == material][cross_correlation_var].values.astype(np.float))
                    logger.info(f'Plotting cross-correlation plots using {cross_correlation_var}')
                    ax.semilogy()
                    ax.set_ylabel(config.borehole_processing.make_material_plots.plot_cross_correlation.label if config.borehole_processing.make_material_plots.plot_cross_correlation.label else f'{cross_correlation_var} [m^2]')


                else:
                    plot_data[material]['y'].append(self.data[borehole][self.data[borehole]['Material_name'] == material]['Z (mAOD)'].values.astype(np.float))

        for material in plot_data:
            plot_x_values = np.concatenate(plot_data[material]['x']).astype(np.float)
            plot_y_values = np.concatenate(plot_data[material]['y']).astype(np.float)
            ax.scatter(x=plot_x_values,
                       y=plot_y_values,
                       label=material,
                       )
            ax.semilogx()
            if config.borehole_processing.make_material_plots.perm_range:
                plot_range = np.array(config.borehole_processing.make_material_plots.perm_range).astype(np.float)
                if variable != 'Porosity (-)':
                    ax.set_xlim(left=plot_range[0],
                                right=plot_range[1],
                                )
        ax.legend()
        if variable is 'Porosity (-)':
            ax.set_xlabel('Porosity [-]')
        elif variable is 'kx':
            ax.set_xlabel('Permeability-x $[m^2]$')
        elif variable is 'ky':
            ax.set_xlabel('Permeability-y $[m^2]$')
        elif variable is 'kz':
            ax.set_xlabel('Permeability-z $[m^2]$')
        elif variable is 'kmean':
            ax.set_xlabel('Geometric mean of permeability $[m^2]$')
        else:
            ax.set_xlabel(f"{variable}")
        plt.savefig(utils.get_output_path() / f'{variable.replace(" ", "_")}_plot.png')
        return fig, ax



import unittest
import numpy.testing as nptest
import numpy as np
from pydelling.utils.geometry import Line, Plane


class LineCase(unittest.TestCase):
    def test_line_init(self):
        line = Line(np.array([0,0,0]), np.array([1,1,0]))
        nptest.assert_array_almost_equal(line.p, np.array([0, 0, 0]))
        nptest.assert_array_almost_equal(line.direction_vector, np.array([np.sqrt(2) / 2.0,np.sqrt(2) / 2.0, 0]))
        nptest.assert_array_equal(line.p, np.array([0, 0, 0]))

    def test_parallel(self):
        line_1 = Line(np.array([0,0,0]), np.array([1,1,0]))
        line_2 = Line(np.array([0,0,0]), np.array([2,2,0]))
        self.assertTrue(line_1.is_parallel(line_2))

    def test_not_parallel(self):
        line_1 = Line(np.array([0,0,0]), np.array([1,1,0]))
        line_2 = Line(np.array([0,0,0]), np.array([3,2,0]))
        self.assertFalse(line_1.is_parallel(line_2))

    def test_intersection_1(self):
        line_1 = Line(np.array([0,0,0]), np.array([-1,0,0]))
        line_2 = Line(np.array([-1,2,0]), np.array([-1,1,0]))
        intersection_point = line_1.intersect(line_2)
        nptest.assert_array_almost_equal(intersection_point, np.array([-1,0,0]))

    def test_intersection_2(self):
        line_1 = Line(np.array([0,0,0]), np.array([2,2,0]))
        line_2 = Line(np.array([2,0,0]), direction_vector=np.array([-1,1,0]))
        intersection_point = line_1.intersect(line_2)
        nptest.assert_array_almost_equal(intersection_point, np.array([1,1,0]))

    def test_intersection_3(self):
        line_1 = Line(np.array([0,0,0]), np.array([0,1,0]))
        line_2 = Line(np.array([0,-1,2]), np.array([0,-1,1]))
        intersection_point = line_1.intersect(line_2)
        nptest.assert_array_almost_equal(intersection_point, np.array([0,-1,0]))

    def test_intersection_4(self):
        line_1 = Line([1, 0, 0], direction_vector=[-1, 0, 1])
        line_2 = Line(p1=[1, 1, 0], p2=[-1, -1, 2])
        line_1_line_2_intersection = line_1.intersect(line_2)
        line_2_line_1_intersection = line_2.intersect(line_1)
        nptest.assert_array_almost_equal(line_2_line_1_intersection, np.array([0, 0, 1]))
        nptest.assert_array_almost_equal(line_1_line_2_intersection, line_2_line_1_intersection)

    def test_intersection_5(self):
        import time
        line_1 = Line([1, 0, 0], direction_vector=[-1, 0, 1])
        line_2 = Line(p1=[1, 1, 0], p2=[-1, -1, 2])

        time_1 = time.time()
        for i in range(1000):
            line_1_line_2_intersection = line_1.intersect(line_2)
        time_2 = time.time()
        delta_time = time_2 - time_1
        self.assertLess(delta_time, 0.5)

    def test_intersect_plane_line(self):
        line = Line(np.array([0,0,0]), np.array([3,0,0]))
        plane = Plane(point=[2.0, 0.0, 0.0], normal=[1.0, 0.0, 0.0])
        intersection_point = line.intersect(plane)
        nptest.assert_array_almost_equal(intersection_point, np.array([2.0, 0.0, 0.0]))


if __name__ == '__main__':
    unittest.main()


from itertools import combinations
from typing import *

import numpy as np

from pydelling.preprocessing.dfn_preprocessor.Fracture import Fracture
from pydelling.utils.geometry import Point, Plane, Line
from pydelling.utils.geometry_utils import filter_unique_points
from .BaseAbstractMeshObject import BaseAbstractMeshObject
from .BaseFace import BaseFace
from scipy.spatial import Delaunay


class BaseElement(BaseAbstractMeshObject):
    local_id = 0
    eps = 1e-2
    eps_zero = 1E-4
    _edge_lines = None
    __slots__ = ['node_ids', 'node_coords']

    def __init__(self, node_ids, node_coords, centroid_coords=None, local_id=None):
        self.nodes: np.ndarray = np.array(node_ids)  # Node id set
        self.coords: np.ndarray = np.array(node_coords)  # Coordinates of each node
        if centroid_coords is None:
            self.centroid = self.compute_centroid()
            self.centroid_coords = self.centroid
        else:
            self.centroid = np.array(centroid_coords)
            self.centroid_coords = self.centroid
        self.local_id = local_id if local_id is not None else BaseElement.local_id
        if local_id is None:
            BaseElement.local_id += 1
        self.faces: Dict[str, BaseFace] = {}  # Faces of an element
        self.type = None
        self.meshio_type = None
        self.associated_fractures = {}
        self.associated_faults = {}
        self.total_fracture_volume = 0
        self.area = 0
        self.is_strange = 0.0

    def __repr__(self):
        return f"{self.type} {self.local_id}"

    def print_element_info(self):
        print("### Element info ###")
        print(f"Element ID: {self.local_id}")
        print(f"Number of nodes: {self.n_nodes}")
        print(f"Element type: {self.type}")
        print(f"Node list: {self.nodes}")
        print("### End element info ###")

    def print_face_info(self):
        print("### Face info ###")
        for face in self.faces:
            print(f"{face}: {self.faces[face].nodes}")
        print("### End face info ###")

    def intersect_faces_with_plane(self, plane: Plane):
        """Returns the intersection of a face with a plane

        Args:
            face: Face to intersect
            plane: Plane to intersect with

        Returns: Intersection points
        """
        intersected_lines = []
        intersected_points = []

        for face in self.faces:
            intersection = self.faces[face].intersect_with_plane(plane)
            if intersection:
                intersected_lines.append(intersection)


        # Intersect the lines within themselves
        intersected_points = list(self._full_line_intersections(intersected_lines=intersected_lines))
        # Check what happens with the fracture points

        # Check if the intersected points are inside the element
        intersected_inside_points = []
        for point in intersected_points:
            if self.contains(point):
                intersected_inside_points.append(point)
        intersected_points = intersected_inside_points.copy()

        return intersected_points


    def intersect_with_fracture(self, fracture: Fracture, export_all_points=False):
        """Intersects an element with a fracture"""
        intersected_lines = []
        intersected_points = []
        final_points = []
        # First thing, check if the fracture is inside the element
        for corner in fracture.corners:
            if self.contains(corner):
                final_points.append(corner)

        if len(final_points) == fracture.n_side_points:
            final_points = [Point(point) for point in final_points]
            return final_points

        for face in self.faces:
            # intersection = self.faces[face].intersect_with_plane(fracture.plane)
            for corner_line in fracture.corner_lines:
                line_intersection = corner_line.intersect(self.faces[face].plane)
                if line_intersection is not None:
                    # if self.contains(line_intersection):
                    intersected_points.append(line_intersection)
            # if intersection:
            #     intersected_lines.append(intersection)
        # Intersect the element edges with the fracture plane
        for edge in self.edge_lines:
            intersection = edge.intersect(fracture.plane)
            edge_intersections = []
            if export_all_points:
                intersection = edge.intersect(fracture.plane)
                edge_intersections.append(intersection)
                with open('custom_edge_intersections.txt', 'a') as f:
                    import csv
                    writer = csv.writer(f)
                    for point in edge_intersections:
                        writer.writerow([point[0], point[1], point[2]])

            if intersection is not None:
                intersected_points.append(intersection)

        # Intersect the lines within themselves
        # intersected_points += list(self._full_line_intersections(intersected_lines=intersected_lines))
        if export_all_points:
            import csv
            with open('intersected_points_all.csv', 'w') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerows(intersected_points)
            self.to_obj('element.obj')
        # Check if the intersected points are inside the element
        intersected_inside_points = []
        for point in intersected_points:
            if self.contains(point):
                intersected_inside_points.append(point)

            # if self.on_face(point):
            #     intersected_inside_points.append(point)
        intersected_points = intersected_inside_points.copy()
        if export_all_points:
            with open('intersected_points_inside.csv', 'w') as csvfile:
                import csv
                writer = csv.writer(csvfile)
                writer.writerows(intersected_inside_points)


        # Test algorithm that checks if a point is contained in the fracture
        for point in intersected_points:
            if fracture.contains(point):
                final_points.append(point)
        if export_all_points:
            print(len(final_points))
            print(final_points)
        # Check if any fracture edge is inside the element. If so, add that as intersection point
        # for corner in fracture.corners:
        #     if self.contains(corner):
        #         final_points.append(corner)

        # final_points = intersected_points

        final_points = filter_unique_points(final_points)
        final_points = [Point(point) for point in final_points]

        return final_points

    def _full_line_intersections(self, intersected_lines: List[Line]) -> List:
        """Intersects a list of lines with each other"""
        intersected_points = []
        line_combination = list(combinations(intersected_lines, 2))
        for line_pair in line_combination:
            line_1: Line = line_pair[0]
            line_2: Line = line_pair[1]
            intersection = line_1.intersect(line_2)
            if intersection is not None:
                intersected_points.append(intersection)
        return intersected_points

    def contains(self, point: np.ndarray or Point, sign=1.0) -> bool:
        """Checks if a point is inside the element"""
        # self.plot_normal_vectors()
        # contains = Delaunay(self.coords).find_simplex(point) >= 0
        # return contains
        initial_sign = None
        for face in self.faces:
            face_centroid = self.faces[face].centroid
            vec = point - face_centroid
            norm_vec = vec / np.linalg.norm(vec)
            dot_plane_point = np.dot(self.faces[face].unit_normal_vector, norm_vec)
            # add a tolerance to the dot product
            if abs(dot_plane_point) < self.eps:
                # Set the same sign as initial_sign
                if initial_sign == None:
                    continue
                elif initial_sign == 1.0:
                    dot_plane_point = 1.0
                elif initial_sign == -1.0:
                    dot_plane_point = -1.0
            if initial_sign is None:
                initial_sign = np.sign(dot_plane_point)
                continue
            if np.sign(dot_plane_point) != initial_sign:
                # self.plot_normal_vectors(point=point, value=dot_plane_point, error_face=face_centroid)
                return False
        return True

    def on_face(self, point):
        """Checks if a point is on a face of the element"""
        for face in self.faces:
            face_centroid = self.faces[face].centroid
            vec = point - face_centroid
            norm_vec = vec / np.linalg.norm(vec)
            dot_plane_point = np.abs(np.dot(self.faces[face].unit_normal_vector, norm_vec))
            if dot_plane_point < self.eps_zero:
                return True
        return False

    @property
    def n_nodes(self):
        return len(self.nodes)

    @property
    def edges(self):
        edges_list = []
        for face in self.faces:
            face_edges = self.faces[face].edges
            edges_list.append(face_edges)

        # Delete duplicates
        edge_list_flatten = [sorted(item) for sublist in edges_list for item in sublist]
        edge_list_unique = np.unique(edge_list_flatten, axis=0)
        return edge_list_unique

    @property
    def volume(self):
        """Returns the volume of the hexahedra

        Returns: volume of the hexahedra
        """
        return None

    def get_json(self):
        """Returns a json representation of the element"""
        serialized_associated_fractures = {key: {
            'area': value['area'],
            'volume': value['volume'],
            'fracture': value['fracture'],
        } for key, value in self.associated_fractures.items()}



        json_dict = {
            "type": self.type,
            "local_id": self.local_id,
            "n_nodes": self.n_nodes,
            "nodes": self.nodes.tolist(),
            "associated_fractures": serialized_associated_fractures,
            "associated_faults": self.associated_faults,
            "total_fracture_volume": self.total_fracture_volume,
            "area": self.area,
        }
        return json_dict

    @property
    def edge_lines(self):
        if self._edge_lines is None:
            # Compute the edge lines
            all_edges = []
            all_lines = []
            for face in self.faces:
                face_edges = self.faces[face].edge_vectors
                for edge_id, edge in enumerate(face_edges):
                    if not self.arr_in_seq(-edge, all_edges):
                        all_edges.append(edge)
                        all_lines.append(Line(p1=self.faces[face].coords[edge_id], direction_vector=edge))
            self._edge_lines = all_lines
        return self._edge_lines


    @staticmethod
    def arr_in_seq(arr, seq):
        tp = type(arr)
        return any(isinstance(e, tp) and np.array_equiv(e, arr) for e in seq)

    @property
    def local_face_nodes(self) -> dict:
        """Returns the local node ids for each face"""
        return {}

    def to_obj(self, filename: str):
        """Exports the element to an obj file"""
        with open(filename, 'w') as f:
            f.write('# OBJ file\n')
            f.write('# Created by pydelling\n')
            f.write('# vertices\n')
            for coord in self.coords:
                f.write('v {} {} {}\n'.format(coord[0], coord[1], coord[2]))
            f.write('# faces\n')
            for face_name in self.local_face_nodes:
                local_ids = self.local_face_nodes[face_name]
                f.write('f ')
                for local_id in local_ids:
                    f.write(f'{local_id + 1} ')
                f.write('\n')


    def compute_centroid(self):
        """
        Computes the centroid of a general polyhedra
        :return: centroid of the polyhedron
        """
        return np.mean(self.coords, axis=0)

    def plot_normal_vectors(self, point: Point=None, value=None, error_face=None):
        """Plots the normal vectors of the faces"""
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d.art3d import Poly3DCollection
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for face in self.faces:
            face_centroid = self.faces[face].centroid
            ax.quiver(face_centroid[0], face_centroid[1], face_centroid[2], self.faces[face].unit_normal_vector[0],
                      self.faces[face].unit_normal_vector[1], self.faces[face].unit_normal_vector[2], length=0.5,
                      normalize=True)
        # Add transparent faces
        for face in self.faces:
            face_coords = self.faces[face].coords
            face_coords = np.array(face_coords)
            ax.add_collection3d(
                Poly3DCollection([face_coords], facecolors='b', alpha=0.1, linewidths=1, edgecolors='k'))


        if point is not None:
            ax.scatter(point[0], point[1], point[2], color='r')

        if value is not None:
            ax.text(point[0], point[1], point[2], value)

        if error_face is not None:
            ax.scatter(error_face[0], error_face[1], error_face[2], color='g')
        # Set aspect ratio
        limits = np.array([ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()])
        origin = np.mean(limits, axis=1)
        radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
        ax.set_xlim3d([origin[0] - radius, origin[0] + radius])
        ax.set_ylim3d([origin[1] - radius, origin[1] + radius])
        ax.set_zlim3d([origin[2] - radius, origin[2] + radius])

        plt.show()



class _AbstractGidObject(object):
    """
    This class contains basic function definitions of the different GiD objects
    """
    def create_gid_bash(self) -> str:
        """
        This method should create a string that contains the bash code to generate the object in GiD
        Returns:
            Bash string
        """
        return ''



import pandas as pd

from pydelling.utils.decorators import set_run


class BasePreprocessing:
    def __init__(self, data: pd.DataFrame = None, filename=None):
        if data is not None:
            self.data = data
        elif filename:
            self.data = pd.read_csv(filename)

    @set_run
    def run(self):
        """
        This method should be implemented for each preprocessing class. It should run the preprocessing algorithm.
        Returns: the result of the preprocessing algorithm
        """
        return 1


from unittest import TestCase
from pydelling.managers import BaseStudy, BaseCallback, BaseManager
from pydelling.utils.configuration_utils import test_data_path


class TestBaseStudy(TestCase):
    def setUp(self) -> None:
        self.base_manager = BaseStudy(str(test_data_path() / 'test_manager.in'))

    def test_replace_tag(self):
        regions = self.base_manager._find_tags('region')
        self.base_manager._replace_line(regions[0], ['REGION', 'pepe'])
        self.assertEqual(self.base_manager._get_line(regions[0]), 'REGION pepe')

    def test_to_file(self):
        self.base_manager.to_file(output_folder='test_folder')
        self.assertTrue(True)

    def test_callback(self):
        BaseManager.__abstractmethods__ = set()
        BaseCallback.__abstractmethods__ = set()

        dummy_study = BaseStudy(str(test_data_path() / 'test_manager.in'))
        dummy_study.add_callback(BaseCallback, kind='pre')

        dummy_manager = BaseManager()
        dummy_manager.add_study(dummy_study)
        dummy_manager.run()
        self.assertEqual(dummy_study.callbacks[0].kind, 'pre')
        self.assertEqual(dummy_study.callbacks[0].study, dummy_study)
        self.assertEqual(dummy_study.callbacks[0].manager, dummy_manager)
        self.assertEqual(dummy_study.callbacks[0].is_run, True)






